@@ -11,13 +11,14 @@ import math
 import numpy as np
 import sys
 
+import psutil
 import requests
 from PIL import Image
 
 from aetros.utils import get_option
 from network import ensure_dir
 
-from threading import Thread
+from threading import Thread, Lock
 from Queue import Queue
 
 def download_image(url, path):
@@ -25,7 +26,7 @@ def download_image(url, path):
         return True
 
     try:
-        r = requests.get(url, stream=True)
+        r = requests.get(url, stream=True, timeout=9)
         if r.status_code == 200:
             with open(path, 'wb') as f:
                 r.raw.decode_content = True
@@ -64,7 +65,7 @@ class ImageDownloaderWorker(Thread):
             return
 
         local_name = image['id']
-        local_image_path = '%s/%s' % (self.trainer.jobModel.get_dataset_downloads_dir(self.dataset), local_name)
+        local_image_path = '%s/%s' % (self.trainer.job_model.get_dataset_downloads_dir(self.dataset), local_name)
 
         try:
             ensure_dir(os.path.dirname(local_image_path))
@@ -93,51 +94,14 @@ class ImageDownloaderWorker(Thread):
         self.images[image['id']] = local_image_path
         self.trainer.set_status('LOAD IMAGE %d/%d' % (len(self.images), self.max))
 
-def read_image(path, size, target_shape, grayscale=False, dim_ordering=None):
-    if dim_ordering is None:
-        import keras.backend
-        dim_ordering = keras.backend.image_dim_ordering()
-
-    try:
-        image = Image.open(path)
-    except:
-        return None
-
-    # do not only resize, maintain aspect ratio
-    image = image.resize(size, Image.ANTIALIAS)
-
-    if grayscale:
-        image = image.convert("L")
-    else:
-        image = image.convert("RGB")
-
-    image = np.asarray(image, dtype='float32')
-
-    if len(target_shape) > 1:
-        # RGB: height, width, channel -> channel, height, width
-        if grayscale:
-            if dim_ordering == 'th':
-                image = image.reshape((1, size[0], size[0]))
-            else:
-                image = image.reshape((size[0], size[0], 1))
-        else:
-            if dim_ordering == 'th':
-                image = image.transpose(2, 0, 1)
-    else:
-        # L: height, width => height*width
-        image = image.reshape(size[0] * size[1])
-
-    return image
-
 # class to read all images in the ram at once
 class ImageReadWorker(Thread):
-    def __init__(self, q, trainer, path, grayscale, size, images, controller):
+    def __init__(self, q, job_model, input_node, path, images, controller):
         Thread.__init__(self)
         self.q = q
-        self.trainer = trainer
+        self.job_model = job_model
+        self.input_node = input_node
         self.path = path
-        self.grayscale = grayscale
-        self.size = size
         self.images = images
         self.controller = controller
 
@@ -151,26 +115,25 @@ class ImageReadWorker(Thread):
 
         try:
             if os.path.isfile(path):
-                image = read_image(path, self.size, self.trainer.input_shape, grayscale=self.grayscale)
+                input = self.job_model.convert_file_to_input_node(path, self.input_node)
             else:
                 return
 
-            if image is None:
-                return
-
-            self.images.append([image, validation, category_dir])
+            if input is not None and len(input) > 0:
+                self.images.append([input, validation, category_dir])
         except IOError as e:
-            self.trainer.logger.write('Could not open %s due to %s' % (path, e.message))
+            print('Could not open %s due to %s' % (path, e.message))
             return
 
 class InMemoryDataGenerator():
-    def __init__(self, datagen, images, categories_count, batch_size):
+    def __init__(self, datagen, images, classes_count, batch_size):
         self.index = -1
         self.datagen = datagen
         self.images = images
         random.shuffle(self.images)
+        self.lock = Lock()
 
-        self.categories_count = categories_count
+        self.classes_count = classes_count
         self.batch_size = batch_size
 
     def __iter__(self):
@@ -182,11 +145,12 @@ class InMemoryDataGenerator():
 
         for i in range(self.batch_size):
 
-            self.index += 1
-            # reset iterator if necessary
-            if self.index == len(self.images):
-                self.index = 0
-                random.shuffle(self.images)
+            with self.lock:
+                self.index += 1
+                # reset iterator if necessary
+                if self.index == len(self.images):
+                    self.index = 0
+                    random.shuffle(self.images)
 
             image, class_idx = self.images[self.index]
             image = np.copy(image) # we need to copy it, otherwise we'd operate on the same object again and again
@@ -194,42 +158,27 @@ class InMemoryDataGenerator():
             if self.datagen is not None:
                 image = self.datagen.random_transform(image)
                 image = self.datagen.standardize(image)
-            else:
-                image /= 255.0
 
             batch_x.append(image)
 
-            y = np.zeros((self.categories_count,), dtype='float32')
+            y = np.zeros((self.classes_count,), dtype='float32')
             y[class_idx] = 1.
             batch_y.append(y)
 
         return np.array(batch_x), np.array(batch_y)
 
 
-def read_images_in_memory(job_config, dataset, node, trainer):
+def read_images_in_memory(job_model, dataset, node, trainer):
     """
     Reads all images into memory and applies augmentation if enabled
     """
-    from keras.utils import np_utils
-    concurrent = 6
+    concurrent = psutil.cpu_count()
 
     dataset_config = dataset['config']
     controller = {'running': True}
     config = dataset['config']
     q = Queue(concurrent)
 
-    size = (node['width'], node['height'])
-
-    if node['inputType'] == 'image':
-        trainer.input_shape = (1, size[0], size[1])
-        grayscale = True
-    elif node['inputType'] == 'image_rgb':
-        grayscale = False
-        trainer.input_shape = (3, size[0], size[1])
-    else:
-        trainer.input_shape = (size[0] * size[1],)
-        grayscale = True
-
     result = {
         'X_train': [],
         'Y_train': [],
@@ -240,17 +189,17 @@ def read_images_in_memory(job_config, dataset, node, trainer):
     images = []
     max = 0
 
-    path = trainer.jobModel.get_dataset_downloads_dir(dataset)
+    path = trainer.job_model.get_dataset_downloads_dir(dataset)
     if 'path' in dataset['config']:
         path = dataset['config']['path']
 
-    categories_count = 0
+    classes_count = 0
     category_map = {}
-    categories = {}
+    classes = []
 
     try:
         for i in range(concurrent):
-            t = ImageReadWorker(q, trainer, path, grayscale, size, images, controller)
+            t = ImageReadWorker(q, job_model, node, path, images, controller)
             t.daemon = True
             t.start()
 
@@ -260,12 +209,16 @@ def read_images_in_memory(job_config, dataset, node, trainer):
                     if os.path.isdir(path+'/'+validation_or_training+'/'+category_name):
 
                         if category_name not in category_map:
-                            category_map[category_name] = categories_count
-                            if 'category_' in category_name:
-                                category_map[category_name] = int(category_name.replace('category_', ''))
+                            category_map[category_name] = classes_count
+                            if 'classes' in dataset_config and 'category_' in category_name:
+                                category_idx = int(category_name.replace('category_', ''))
+                                category_map[category_name] = category_idx
+                                target_category = dataset_config['classes'][category_idx]
+                                classes.append(target_category['title'] or 'Class %s' % (category_idx, ))
+                            else:
+                                classes.append(category_name)
 
-                            categories[categories_count] = category_name
-                            categories_count += 1
+                            classes_count += 1
 
                         for id in os.listdir(path+'/'+validation_or_training+'/'+category_name):
                             file_path = os.path.join(path, validation_or_training, category_name, id)
@@ -289,21 +242,25 @@ def read_images_in_memory(job_config, dataset, node, trainer):
         augmentation = bool(get_option(dataset_config, 'augmentation', False))
         if augmentation:
             train_datagen = get_image_data_augmentor_from_dataset(dataset)
-        train = InMemoryDataGenerator(train_datagen, train_images, categories_count, job_config['settings']['batchSize'])
 
-        test = InMemoryDataGenerator(None, test_images, categories_count, job_config['settings']['batchSize'])
+        train = InMemoryDataGenerator(train_datagen, train_images, classes_count, job_model.job['config']['settings']['batchSize'])
 
-        print ("Found %d categories, %d images (%d in training [%saugmented], %d in validation). Read all images into memory from %s" %
-               (categories_count, max, len(train_images), 'not ' if augmentation is False else '', len(test_images), path))
-        pprint(category_map)
+        test = InMemoryDataGenerator(None, test_images, classes_count, job_model.job['config']['settings']['batchSize'])
 
         nb_sample = len(train_images)
-        trainer.samples_per_epoch = nb_sample
+        trainer.set_generator_training_nb(nb_sample)
+        trainer.set_generator_validation_nb(len(test_images))
 
-        trainer.nb_val_samples = len(test_images)
-        trainer.output_size = categories_count
-        trainer.set_job_info('categories', categories)
-        pprint(categories)
+        print ("Found %d classes, %d images (%d in training [%saugmented], %d in validation). Read all images into memory from %s" %
+               (classes_count, max, len(train_images), 'not ' if augmentation is False else '', len(test_images), path))
+
+        if classes_count == 0:
+            print("Could not find any classes. Does the directory contains images?")
+            sys.exit(1)
+
+        trainer.output_size = classes_count
+        trainer.set_job_info('classes', classes)
+        trainer.classes = classes
 
         result['X_train'] = train
         result['Y_train'] = train
@@ -327,7 +284,6 @@ def get_image_data_augmentor_from_dataset(dataset):
     augRotationRange = float(get_option(dataset_config, 'augRotationRange', 0.2))
 
     return ImageDataGenerator(
-            rescale=1./255,
             rotation_range=augRotationRange,
             shear_range=augShearRange,
             zoom_range=augZoomRange,
@@ -335,19 +291,13 @@ def get_image_data_augmentor_from_dataset(dataset):
             vertical_flip=augVerticalFlip
     )
 
-def read_images_keras_generator(job_config, dataset, node, trainer):
+def read_images_keras_generator(job_model, dataset, node, trainer):
     from keras.preprocessing.image import ImageDataGenerator
 
-    size = (node['width'], node['height'])
+    size = (int(node['width']), int(node['height']))
 
+    grayscale = False
     if node['inputType'] == 'image':
-        trainer.input_shape = (1, size[0], size[1])
-        grayscale = True
-    elif node['inputType'] == 'image_rgb':
-        grayscale = False
-        trainer.input_shape = (3, size[0], size[1])
-    else:
-        trainer.input_shape = (size[0] * size[1],)
         grayscale = True
 
     dataset_config = dataset['config']
@@ -357,26 +307,34 @@ def read_images_keras_generator(job_config, dataset, node, trainer):
 
     if augmentation:
         train_datagen = get_image_data_augmentor_from_dataset(dataset)
+
+        if 'imageScale' not in node:
+            node['imageScale'] = 255
+
+        if float(node['imageScale']) > 0:
+            train_datagen.rescale = 1.0 / float(node['imageScale'])
     else:
         train_datagen = ImageDataGenerator()
 
     train_generator = train_datagen.flow_from_directory(
             directory=os.path.join(dataset_config['path'], 'training'),
             target_size=size,
-            batch_size=job_config['settings']['batchSize'],
+            batch_size=job_model.job['config']['settings']['batchSize'],
             color_mode='grayscale' if grayscale is True else 'rgb',
             class_mode='categorical')
 
-    categories = {}
+    classes = []
     for folderName, outputNeuron in train_generator.class_indices.iteritems():
         if dataset['type'] == 'images_search' or dataset['type'] == 'images_upload':
             category_idx = int(folderName.replace('category_', ''))
-            target_category = dataset_config['categories'][category_idx]
-            categories[outputNeuron] = target_category['title'] or 'Category %s' % (category_idx, )
+            target_category = dataset_config['classes'][category_idx]
+            classes.append(target_category['title'] or 'Category %s' % (category_idx, ))
         else:
-            categories[outputNeuron] = folderName
+            classes.append(folderName)
+
+    trainer.set_job_info('classes', classes)
+    trainer.classes = classes
 
-    trainer.set_job_info('categories', categories)
     trainer.output_size = train_generator.nb_class
 
     # ensure_dir(dataset_config['path'] + '/preview')
@@ -386,18 +344,22 @@ def read_images_keras_generator(job_config, dataset, node, trainer):
             directory=os.path.join(dataset_config['path'], 'validation'),
             # save_to_dir=dataset_config['path'] + '/preview',
             target_size=size,
-            batch_size=job_config['settings']['batchSize'],
+            batch_size=trainer.get_batch_size(),
             color_mode='grayscale' if grayscale is True else 'rgb',
             class_mode='categorical')
 
-    trainer.nb_val_samples = validation_generator.nb_sample
-    trainer.samples_per_epoch = train_generator.nb_sample
+    trainer.set_generator_validation_nb(validation_generator.nb_sample)
+    trainer.set_generator_training_nb(train_generator.nb_sample)
 
-    print ("Found %d categories, %d images (%d in training [%saugmented], %d in validation) in %s " %
-           (len(categories), trainer.nb_val_samples+trainer.samples_per_epoch, trainer.samples_per_epoch, 'not ' if augmentation is False else '', trainer.nb_val_samples, dataset_config['path']))
+    print ("Found %d classes, %d images (%d in training [%saugmented], %d in validation) in %s " %
+           (len(classes), validation_generator.nb_sample+train_generator.nb_sample, train_generator.nb_sample, 'not ' if augmentation is False else '', validation_generator.nb_sample, dataset_config['path']))
+
+    if trainer.output_size == 0:
+        print("Could not find any classes. Does the directory contains images?")
+        sys.exit(1)
 
     pprint(train_generator.class_indices)
-    pprint(categories)
+    pprint(classes)
 
     return {
         'X_train': train_generator,
@@ -406,39 +368,29 @@ def read_images_keras_generator(job_config, dataset, node, trainer):
         'Y_test': validation_generator,
     }
 
-def get_images(job_config, dataset, node, trainer):
+def get_images(job_model, dataset, node, trainer):
     concurrent = 15
 
     q = Queue(concurrent)
     config = dataset['config']
 
-    dir = trainer.jobModel.get_dataset_downloads_dir(dataset)
+    dir = trainer.job_model.get_dataset_downloads_dir(dataset)
 
     ensure_dir(dir)
-
-    categories = config['categories']
-
-    size = (node['width'], node['height'])
+    classes = config['classes']
 
     trainer.set_status('PREPARE_IMAGES')
 
-    if node['inputType'] == 'image':
-        trainer.input_shape = (1, size[0], size[1])
-    elif node['inputType'] == 'image_rgb':
-        trainer.input_shape = (3, size[0], size[1])
-    else:
-        trainer.input_shape = (size[0] * size[1],)
-
     max = 0
     images = {}
 
-    dataset_path = trainer.jobModel.get_dataset_downloads_dir(dataset)
+    dataset_path = trainer.job_model.get_dataset_downloads_dir(dataset)
     meta_information_file = dataset_path + '/meta.json'
 
-    categories_changed = False
+    classes_changed = False
     config_changed = False
     had_previous = False
-    categories_md5 = hashlib.md5(json.dumps(categories)).hexdigest()
+    classes_md5 = hashlib.md5(json.dumps(classes)).hexdigest()
 
     validationFactor = 0.2
 
@@ -448,8 +400,8 @@ def get_images(job_config, dataset, node, trainer):
                 meta = json.load(f)
                 if meta:
                     had_previous = True
-                    if meta['categories_md5'] != categories_md5:
-                        categories_changed = True
+                    if 'classes_md5' in meta and meta['classes_md5'] != classes_md5:
+                        classes_changed = True
 
                     trigger_changed = ['resize', 'resizeWidth', 'resizeHeight', 'resizeCompression']
                     for i in trigger_changed:
@@ -460,13 +412,13 @@ def get_images(job_config, dataset, node, trainer):
         else:
             config_changed = True
 
-    need_download = categories_changed or config_changed
+    need_download = classes_changed or config_changed
 
     if need_download:
         if had_previous:
             print "Reset dataset and re-download images to " + dir
-            if categories_changed:
-                print (" .. because categories changed")
+            if classes_changed:
+                print (" .. because classes changed")
             if config_changed:
                 print (" .. because settings changed")
         else:
@@ -482,7 +434,7 @@ def get_images(job_config, dataset, node, trainer):
 
         controller = {'running': True}
         try:
-            for category in categories:
+            for category in classes:
                 max += len(category['images'])
 
             for i in range(concurrent):
@@ -490,7 +442,7 @@ def get_images(job_config, dataset, node, trainer):
                 t.daemon = True
                 t.start()
 
-            for category_idx, category in enumerate(categories):
+            for category_idx, category in enumerate(classes):
                 for image in category['images']:
                     q.put([image, category_idx])
 
@@ -503,7 +455,7 @@ def get_images(job_config, dataset, node, trainer):
                     ensure_dir(os.path.dirname(target_path))
                     os.rename(images[image['id']], target_path)
 
-            for category_idx, category in enumerate(categories):
+            for category_idx, category in enumerate(classes):
                 random.shuffle(category['images'])
                 position = int(math.ceil(len(category['images']) * validationFactor))
 
@@ -520,8 +472,8 @@ def get_images(job_config, dataset, node, trainer):
 
             with open(meta_information_file, 'w') as f:
                 meta = {
-                    'loaded_at': categories_md5,
-                    'categories_md5': categories_md5,
+                    'loaded_at': classes_md5,
+                    'classes_md5': classes_md5,
                     'config': config
                 }
                 json.dump(meta, f)
@@ -533,7 +485,7 @@ def get_images(job_config, dataset, node, trainer):
         print "Downloaded images up2date in " + dir
         print " - Remove this directory if you want to re-download all images of your dataset and re-shuffle training/validation images."
 
-    trainer.output_size = len(categories)
+    trainer.output_size = len(classes)
     trainer.set_status('LOAD IMAGE DONE')
 
     # change to type local_images
@@ -543,6 +495,6 @@ def get_images(job_config, dataset, node, trainer):
     all_memory = get_option(dataset['config'], 'allMemory', False, 'bool')
 
     if all_memory:
-        return read_images_in_memory(job_config, dataset_transformed, node, trainer)
+        return read_images_in_memory(job_model, dataset_transformed, node, trainer)
     else:
-        return read_images_keras_generator(job_config, dataset_transformed, node, trainer)
\ No newline at end of file
+        return read_images_keras_generator(job_model, dataset_transformed, node, trainer)
\ No newline at end of file
